//******************************************************************
//  Setup.cpp
//
//  This file was originally downloaded from the BCPPWindows forum on
//  Compuserve.  There was no indication as to the original author or
//  details on licensing.  If by amending and re-releasing this code
//  I have broken any license terms then please contact me.
//
//  I have amended this code to provide a more "professional" user
//  interface.  I have made no attempt to pretty the code or structure it in
//  a more rational way.  I place my, messy, amendments into the public domain.
//
//  Note that my tabs are set to 2 spaces.
//
//  Possible improvements: (It currently does what I want so I'm done with it!)
//    Ability to install to multiple directories.
//    Allow more than one install source disk.	
//    Thermometer style display for progress made.
//
//  Stewart McSporran, May 1993
//  Compuserve 100064,773 
//
//  Thanks to the unknown original author.
//******************************************************************
//  FURTHER AMENDMENTS:
//
//******************************************************************
#define WIN31
#include <Owl.h>
#include <Bwcc.h>
#include <dir.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include <stdio.h>
#include <ctype.h>
#include <lzexpand.h>
#include <dde.h>
#include <static.h>
#pragma hdrstop;
#include "Setup.h"


//***
//* Globals
//***

MainDialog *GMainDialog;   // global pointer to mainwindow instance
BOOL GroupCreated = FALSE;
char szGroupName[80];			 // program manager group
char szFromDrive[MAXDRIVE];//drive SETUP.EXE was run from
char szFromDir[MAXPATH];   //path SETUP.EXE was run from
char szIniFile[MAXPATH];   //full pathcpec of ini file
char szToDrive[MAXDRIVE];  //drive to install to
char szToDir[MAXDIR];	   //path to install to
char szAppName[255];	   //name of application as specified in ini file
char szCommand[512];	   //global buffer used for DDE commands
int  iSUCCESS = FALSE;     // if TRUE, proper messagebox is displayed.
char szWindowTitle[255];   // used to find > one instance of program

//********************************************
//* Assist functions
//*********************************************
int lstrichr(LPSTR string,char c)
// looks thru string for char c, ignoring case
{
	char c2=c;
	toupper(c2);
	for (int i=0;string[i];i++)
		if (string[i]==c2 || string[i]+32==c2)
			return string[i];
	return FALSE;
}

void HourGlass(int iOn)
// turns hourglass cursor on or off
{
	static HCURSOR hcurSave;
    if (iOn)
		hcurSave = SetCursor(LoadCursor(NULL, IDC_WAIT));
	else
		SetCursor(hcurSave);
}

void Error(LPSTR msg)
// displays messagebox with 'Error' title and stop sign icon
{
	MessageBeep(MB_ICONSTOP);
	BWCCMessageBox(GetActiveWindow(),msg,NULL,MB_ICONSTOP);
}

void CentreDialog(HWND HWindow)
{
	int X = GetSystemMetrics(SM_CXSCREEN) / 2;
	int Y = GetSystemMetrics(SM_CYSCREEN) / 2;
	RECT r;
	GetWindowRect(HWindow,&r);
	SetWindowPos(HWindow,0,X-(r.right-r.left)/2,Y-(r.bottom-r.top)/2,r.right,r.bottom,SWP_NOSIZE|SWP_NOZORDER);
}

// called by the create group dialog box, calls MainDialog members
BOOL CreateProgmanGroup()
{
	return (GMainDialog->InitiateDDE() && GMainDialog->CreateGroup() &&
					GMainDialog->CreateItems() && GMainDialog->TerminateDDE());
}

//***
//* Program entry
//***
#pragma argsused;
int PASCAL WinMain(HANDLE hInstance,HANDLE
                    hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
{
	// store pathspec SETUP.EXE is run from
	char szStartApp[MAXPATH];
	GetModuleFileName(hInstance,szStartApp,MAXPATH-1);
	fnsplit(szStartApp,szFromDrive,szFromDir,NULL,NULL);
    // look for SETUP.INI file
	fnmerge(szIniFile,szFromDrive,szFromDir,"SETUP","INI");
    OFSTRUCT of;
	if (HFILE_ERROR==OpenFile(szIniFile,&of,OF_EXIST)) {
        Error("SETUP.INI file is missing - can't run setup.");
		return 99;
	}
	  // get name of application we are installing from the ini file
    GetPrivateProfileString("Application",
                            "Name",
                            "",
                            szAppName,
                            MAXPATH-1,
							szIniFile);
    // format into a string to plug into window title later 
	  sprintf(szWindowTitle,"Setup - %s",szAppName);
	  // check for program running > once, redirect focus
    if (hPrevInstance) {
		HWND hwnd = FindWindow(NULL,szWindowTitle);
		hwnd= GetLastActivePopup(hwnd);
		BringWindowToTop(hwnd);
		ShowWindow(hwnd,SW_RESTORE);
		return 0;
	}
	else {
		// load LZEXPAND.DLL
		HANDLE hLib = LoadLibrary("lzexpand.dll");
		if(hLib <= 32)
			Error("Can't load lzexpand.dll");
		// load BWCC.DLL
		HINSTANCE hBorLibrary = LoadLibrary("bwcc.dll");
    if((UINT)hBorLibrary <= 32)
		  MessageBox(NULL, "Unable to load Borland Controls", "System Error", MB_OK | MB_ICONHAND);

		// start derived class of TApplication
   	SetupApp SetupApp ("SetupApp",
                        hInstance,
                        hPrevInstance,
                        lpCmdLine,
												nCmdShow);
    SetupApp.Run();
			// program is essentially complete; display message at exit
		char szMsg[255];
		if (iSUCCESS)
    {
			if (GroupCreated)
      {
				sprintf(szMsg,"Setup has installed '%s' on to your system and added it to your Windows desktop.",szAppName);
			}
			else
			{
				sprintf(szMsg,"Setup has installed '%s' on to your system.",szAppName);
			}
		}
		else
    {
			sprintf(szMsg,"'%s' was not installed.",szAppName);
    }

		BWCCMessageBox(GetActiveWindow(),szMsg,"Setup Status",MB_ICONINFORMATION);
    if ((UINT)hLib > 32)
			FreeLibrary(hLib);
		if((UINT)hBorLibrary > 32)
		  FreeLibrary(hBorLibrary);
    return SetupApp.Status;
	}
}
//***************  Status dialog *************
class TStatusDLG:public TDialog{
public:
	TStatusDLG(PTWindowsObject AParent, LPSTR AName):TDialog(AParent,AName){}
  virtual void SetupWindow();
};

void TStatusDLG::SetupWindow()
{
	TDialog::SetupWindow();
	// centre the dialog
	CentreDialog(HWindow);
}

//***************  Create Prog group dialog *************
class TProgManDLG:public TDialog{
public:
  PTStatic pGroupTS;
  BOOL GroupName;
	TProgManDLG(PTWindowsObject AParent, LPSTR AName);
  ~TProgManDLG();
	virtual void SetupWindow();
	virtual void Yes(RTMessage) = [ID_FIRST + IDYES];
	virtual void No(RTMessage) = [ID_FIRST + IDNO];
};

TProgManDLG::TProgManDLG(PTWindowsObject AParent, LPSTR AName):TDialog(AParent,AName)
{
	GetPrivateProfileString("Application","GroupName",
							"",szGroupName,79,szIniFile);
  GroupName = TRUE;
	if (!lstrlen(szGroupName))
	{
		Error("No Group name in SETUP.INI");
		GroupName = FALSE;
	}
	pGroupTS = new TStatic(this,ID_PROGGROUPNAME,40);
}

TProgManDLG::~TProgManDLG()
{
  delete pGroupTS;
}

void TProgManDLG::SetupWindow()
{
	TDialog::SetupWindow();
	// centre the dialog
	CentreDialog(HWindow);
  pGroupTS->SetText(szGroupName);
	if (!GroupName)
		CloseWindow(FALSE);
}

#pragma argsused
void TProgManDLG::Yes(RTMessage Msg)
{
  // user has selected to add to program manager
	if(CreateProgmanGroup())
	{
    GroupCreated = TRUE;
    CloseWindow(TRUE);
	}
	else
	{
    CloseWindow(FALSE);
  }
}

#pragma argsused
void TProgManDLG::No(RTMessage Msg)
{
  // return true otherwise the program will show the "error in installation" dialog
  CloseWindow(TRUE);
}

//*************************
// TApplication descendant - SetupApp
//*************************
void SetupApp::InitMainWindow()
{
	MainWindow = new MainDialog();
  GMainDialog = (MainDialog *)MainWindow;
}
//*******************
// Main Dialog
//*******************
MainDialog::MainDialog():TDialog(NULL,"MAIN")
{
  DDEMessageSent=0; //DDEMessageSent is used to execute a message loop that
										//keeps ProgMan from losing messages due to a full queue
}
void MainDialog::WMInitDialog(RTMessage msg)
{
	TDialog::WMInitDialog(msg);
	char szTemp[MAXPATH];

    // plug default directory from ini file into edit control in dialog
    GetPrivateProfileString("Application",
                            "DefaultPath",
                            "",
                            szTemp,
                            MAXPATH-1,
                            szIniFile);
	SetDlgItemText(HWindow,IDPATH,szTemp);
	SendDlgItemMsg(IDPATH,EM_LIMITTEXT,MAXPATH-2,0L);
	// set window title to include name of application
	SetCaption(szWindowTitle);
}

void MainDialog::SetupWindow()
{
	TDialog::SetupWindow();
	// centre the dialog
	CentreDialog(HWindow);
}

LPSTR MainDialog::GetClassName()
{
  return "MainDlg";
}
void MainDialog::GetWindowClass(WNDCLASS _FAR & aclass)
{
   TDialog::GetWindowClass( aclass );
   //aclass.lpfnWndProc = BWCCDefDlgProc;
   HANDLE h=LoadIcon(GetApplication()->hInstance,"SETUP");
   aclass.hIcon =h;
}

void MainDialog::Ok(RTMessage Msg)
// the OK Button launches the entire application
{
	TProgManDLG *pmStatus;  // program manager install request dialog
	pmStatus = new TProgManDLG(this,"PMSTATUS");
	pmStatus->SetCaption(szWindowTitle);
	if(CheckPathGiven() &&
		CheckDiskSpace() &&
		MakeDirectory()  &&
		CopyFiles() 	 &&
		GetModule()->ExecDialog(pmStatus))  // ask if wants to add prog manager group
  {
		  TDialog::Ok(Msg);
      iSUCCESS=TRUE;
  }
	else
	{
		iSUCCESS=FALSE;
		TDialog::Cancel(Msg);
	}
}
int MainDialog::CheckPathGiven(void)
// parses pathname entered in dialog into drive and path components
// also checks validity by using fnsplit.
{
	char szTemp[MAXPATH];
	GetDlgItemText(HWindow,IDPATH,szTemp,MAXPATH-2);
	// add backslash so fnsplit works properly
	if (szTemp[lstrlen(szTemp)]!='\\')
		lstrcat(szTemp,"\\");
	fnsplit((char*)szTemp,(char*)szToDrive,(char*)szToDir,NULL,NULL);
    if (!lstrlen(szToDrive) || ! lstrlen(szToDir)) {
		Error("Invalid path entered.");
		return FALSE;
	}
    // fall thru to OK
	return TRUE;
}
int MainDialog::CheckDiskSpace(void)
// check disk space available according to value in ini file
{
	char szTemp[20];
	GetPrivateProfileString("Application","SizeNeeded",
                            "",szTemp,19,szIniFile);
	long lSpaceNeeded = atol(szTemp);
    if (lSpaceNeeded==0L) {
        Error("Disk space needed is zero - can't run setup.");
		return FALSE;
    }
    // check disk space on drive specified
	long avail;
	int drive=toupper(szToDrive[0])-'A';
	struct dfree free;
	getdfree(drive+1, &free);
    if (free.df_sclus == 0xFFFF) {
        Error("Error getting disk space");
		return FALSE;
	}
	avail =  (long) free.df_avail *
			(long) free.df_bsec * (long) free.df_sclus;
	if (avail < lSpaceNeeded) {
		char szMsg[255];
		sprintf(szMsg,
		"%ld bytes needed, Drive %c: only has %ld bytes available\n",
		lSpaceNeeded,'A' + drive, avail);
		Error(szMsg);
		return FALSE;
	}
	return TRUE;
}
int MainDialog::MakeDirectory(void)
// create directory specified in ini file
{
	char szTemp[MAXPATH];
    lstrcpy(szTemp,szToDrive);
	lstrcat(szTemp,szToDir);
	szTemp[lstrlen(szTemp)-1]='\0'; //strip last backslash
	mkdir(szTemp); 	//ignore possible error - path may already exist,
					//thats OK
	return TRUE;
}
int MainDialog::CopyFiles(void)
// read all files from ini file and either copy or decompress them
{
	HANDLE h=GlobalAlloc(GHND,10000);
	if (!h) {
		Error("Not enough memory");
		return FALSE;
    }
	LPSTR szBuffer=GlobalLock(h);
	char szLine[255];
	char szFileName[MAXPATH];
	char szDesc[80];
	char szCompressed[10];
	char szCopyOk[10];

	// Display hourglass.
	HourGlass(TRUE);
	// get string that contains all files to copy, seperated
	// by nulls.GetPrivateString will do this if passed NULL in 2nd param
	GetPrivateProfileString("FILES",NULL,"",szBuffer,9999,szIniFile);
	LPSTR p=szBuffer;
	// count # of files - only needed to display % complete
	int iFiles;
	for (iFiles=0;*p;iFiles++)
		p+=lstrlen(p)+1;
	// Display non-modal dialog to track status..
	TStatusDLG *StatDlg=new TStatusDLG(this,"STATUS");
	GetModule()->MakeWindow(StatDlg);
	StatDlg->SetCaption(szWindowTitle);
	StatDlg->Show(SW_SHOW);
	p=szBuffer;
	int iDoneFiles=0;
	int iGo;
	while (*p) {
		// display % complete in status dialog
		// this is purposely displayed ahead of the actual copy, which is
		// not accurate but otherwise the 100% vanishes so
		// quickly that it is confusing to the user
        iDoneFiles++;
		int iPct = (iDoneFiles*100)/iFiles;
		char szMsg[80];
		sprintf(szMsg,"%3.0d %% complete",iPct);
		SetDlgItemText(StatDlg->HWindow,IDPERCENT,szMsg);

		// get info for every file
		GetPrivateProfileString("FILES",p,"",szLine,254,szIniFile);
		lstrcpy(szFileName,p);
		// parse info line into components
		// warning..no error checking done, strings must be correct in
		// .ini file or most likely will cause GPF...
		LPSTR s=strtok(szLine,",");
		lstrcpy(szDesc,s);
		s=strtok(NULL,",");
		lstrcpy(szCompressed,s);
		s=strtok(NULL,",");
		lstrcpy(szCopyOk,s);
		p+=lstrlen(p)+1;

		// update status dialog file descr..
		sprintf(szMsg,"%s",szDesc);
		SetDlgItemText(StatDlg->HWindow,IDFILENAME,szMsg);

		// Do the copy or LzExpand for every file...
		if (lstrichr(szCompressed,'Y'))
			iGo=UnCompress(szFileName,szCopyOk);
		else
			iGo=Copy(szFileName,szCopyOk);
		if (!iGo)
			break;
	}
	// vanish status dialog
	StatDlg->CloseWindow();
    // release memory
	GlobalUnlock(h);
	GlobalFree(h);
	HourGlass(FALSE);
	return iGo;
}
int MainDialog::Copy(LPSTR  filename,LPSTR  copyok)
// perform a low level copy for a file
{
	char szSrc[MAXPATH];
	char szDest[MAXPATH];
	sprintf(szSrc,"%s%s%s",szFromDrive,szFromDir,filename);
	sprintf(szDest,"%s%s%s",szToDrive,szToDir,filename);

	int nDstFile;
	WORD wBytesRead, wDate, wTime;
	OFSTRUCT ofSrc, ofDst;

	// check for overwriting existing files..
	if (!lstrichr(copyok,'Y') &&
		HFILE_ERROR!=OpenFile(szDest, &ofDst, OF_EXIST))
		return TRUE;

	int in = OpenFile(szSrc, &ofSrc, OF_READ);
	if (in == -1) { 
		Error("Cant open input file");
		return FALSE;
    }

	int out = OpenFile(szDest, &ofDst, OF_CREATE | OF_WRITE);
	if (out == -1) {
		Error("Cant create output file");
		return FALSE;
    }

	unsigned int bSize=50000;
	HANDLE h=GlobalAlloc(GHND,bSize);
	if (!h) {
		Error("Not enough memory");
		return FALSE;
	}
	//grab memory, do copy
	LPSTR mem=GlobalLock(h);
	unsigned int Bytes;
	do {
	   Bytes = _lread(in, mem, bSize);
	   _lwrite(out, mem, Bytes);
	} while (Bytes == bSize);

	// Make the dest and source file have same time stamp.
	unsigned int Date,Time;
	_dos_getftime(in, &Date, &Time);
	_dos_setftime(out, Date, Time);
	_lclose(out);
	_lclose(in);
	GlobalUnlock(h);
	GlobalFree(h);
	return TRUE;
}
int MainDialog::UnCompress(LPSTR  filename,LPSTR  copyok)
// call LZE function to decompress file
{
	char szSrc[MAXPATH];
	char szDest[MAXPATH];
	char szUnCompressedFile[MAXFILE];
	char szUnCompressedExt[MAXEXT];
	sprintf(szSrc,"%s%s%s",szFromDrive,szFromDir,filename);
	// get filename without funky underscore in extension
	if (GetExpandedName(szSrc,szDest)==-1) {
		char szMsg[80];
		sprintf(szMsg,"Can't expand file name for %s...\r\nIs it missing?",szSrc);
		Error(szMsg);
		return FALSE;
	}
    // build full target path
	fnsplit(szDest,NULL,NULL,szUnCompressedFile,szUnCompressedExt);
	sprintf(szDest,"%s%s%s%s",szToDrive,szToDir,
							szUnCompressedFile,szUnCompressedExt);
	OFSTRUCT ofSrc, ofDst;
	// check for existing file not allowed to overwrite
	if (! lstrichr(copyok,'Y') &&
		HFILE_ERROR!=OpenFile(szDest, &ofDst, OF_EXIST))
		return TRUE;

	HFILE in = LZOpenFile(szSrc, &ofSrc, OF_READ);
	if (in == -1) {
		Error("Cant open LZE file");
		return FALSE;
	}
	HFILE out = LZOpenFile(szDest, &ofDst, OF_CREATE);
	if (out == -1) {
		Error("Cant create LZE file");
		return FALSE;
	}
	// actual decompression done here
	LZCopy(in,out);
	LZClose(in);
	LZClose(out);
	return TRUE;
}
int MainDialog::InitiateDDE(void)
// starts DDE connection to Program manager
{
	aApp  = GlobalAddAtom( "PROGMAN" );
	aTopic= GlobalAddAtom( "PROGMAN" );
	//these will be deleted at the proper time in HandleDDEAck
	DDEMessageSent=WM_DDE_INITIATE;
	SendMessage( -1, WM_DDE_INITIATE, HWindow, MAKELONG( aApp, aTopic) );
    return TRUE;
}
int MainDialog::CreateGroup(void)
// creates the new PM group by calling the ExecutePMDDE function
{
	// go ahead and create the group, even if it existed before.
  // Presentation manager will simply fail to add another.
	sprintf(szCommand,"[CreateGroup(%s)]",szGroupName);
	ExecutePMDDE(szCommand);
	sprintf(szCommand,"[ShowGroup(%s,1)]",szGroupName);
	ExecutePMDDE(szCommand);
	return TRUE;
}
int MainDialog::CreateItems(void)
// creates the new PM items by calling the ExecutePMDDE function
{

	char szItemExe[MAXPATH];
	char szItemDesc[80];
	char szItemIconFile[MAXPATH];
	char szItemIconNumber[10];
	char szLine[255];
	HANDLE h=GlobalAlloc(GHND,10000);
	if (!h) {
		Error("Not enough memory");
		return FALSE;
	}
    // get list of all items to add to group, delim with NULLS
	LPSTR szBuffer=GlobalLock(h);
	GetPrivateProfileString("ITEMS",NULL,"",szBuffer,9999,szIniFile);
	LPSTR p=szBuffer;
    // look thru list
	while (*p) {
		memset(szItemExe,0,sizeof(szItemExe));
		memset(szItemDesc,0,sizeof(szItemDesc));
		memset(szItemIconFile,0,sizeof(szItemIconFile));
		memset(szItemIconNumber,0,sizeof(szItemIconNumber));
		memset(szLine,0,sizeof(szLine));
		// get info for every item
		GetPrivateProfileString("ITEMS",p,"",szLine,254,szIniFile);
		// build full pathspec for file
		lstrcpy(szItemExe,szToDrive);
		lstrcat(szItemExe,szToDir);
        lstrcat(szItemExe,p);

		LPSTR s=strtok(szLine,",");
		lstrcpy(szItemDesc,s);
		s=strtok(NULL,",");
		lstrcpy(szItemIconFile,s);
		s=strtok(NULL,",");
		lstrcpy(szItemIconNumber,s);
		p+=lstrlen(p)+1;
		if (lstrlen(szItemIconFile)) {
        	// provide path for icon file if not already included
			if (!strchr(szItemIconFile,'\\')) {
            	lstrcpy(szLine,szItemIconFile);
				lstrcpy(szItemIconFile,szToDrive);
				lstrcat(szItemIconFile,szToDir);
				lstrcat(szItemIconFile,szLine);
			}
			if (! lstrlen(szItemIconNumber))
				// provide default icon # 0
				lstrcpy(szItemIconNumber,"0");
			sprintf(szCommand,"[AddItem(%s,%s,%s,%s)]",
			szItemExe,szItemDesc,szItemIconFile,szItemIconNumber);
		}
		else {
        	if (lstrlen(szItemIconNumber))
				sprintf(szCommand,"[AddItem(%s,%s,%s,%s)]",
				szItemExe,szItemDesc,szItemExe,szItemIconNumber);
			else
				sprintf(szCommand,"[AddItem(%s,%s)]",szItemExe,szItemDesc);
		}
		ExecutePMDDE(szCommand);
	}
	// release memory used to read all items in .ini file
	GlobalUnlock(h);
	GlobalFree(h);
	return TRUE;
}
int MainDialog::TerminateDDE(void)
{
	// send message to end communication
	PostMessage( -1, WM_DDE_TERMINATE, HWindow, NULL );
	return TRUE;
}
void MainDialog::HandleDDEAck(RTMessage Msg)
{
	switch(DDEMessageSent) {
		case WM_DDE_INITIATE:
			GlobalDeleteAtom( Msg.LP.Lo );
			GlobalDeleteAtom( Msg.LP.Hi );
            DDEMessageSent=0;
			break;
		case WM_DDE_EXECUTE:
			GlobalFree(hDDE);
			DDEMessageSent=0;
			break;
	}
}
void MainDialog::ExecutePMDDE(LPSTR  string)
{
	LPSTR   lpExecute;
	// allocate atoms for topic and app names
	// grab global memory - Program Manager will delete it..
	hDDE = GlobalAlloc( GMEM_MOVEABLE | GMEM_DDESHARE,
		lstrlen(string)+10 );
	lpExecute = GlobalLock( hDDE );
	lstrcpy( lpExecute, string);
	GlobalUnlock(hDDE);
	DDEMessageSent=WM_DDE_EXECUTE;
	PostMessage( -1, WM_DDE_EXECUTE, HWindow, MAKELONG( 0, hDDE) );
    MSG Msg;
	while (DDEMessageSent) {
		GetMessage(&Msg, HWindow, 0,0);
		TranslateMessage(&Msg);
		DispatchMessage(&Msg);
    }
}

